theory Separation
  imports Semantics HOL.Hilbert_Choice
begin

type_synonym 'a val  = "'a state \<Rightarrow> 'a"
type_synonym 'a pure = "'a state \<Rightarrow> bool"
type_synonym 'a rel  = "'a state \<Rightarrow> 'a state \<Rightarrow> bool"
type_synonym 'a pred = "'a state * 'a heap \<Rightarrow> 'a state * 'a heap \<Rightarrow> bool"

lemma map_add_empty_elim[elim!]:
"(Map.empty = m1 ++ m2) \<Longrightarrow> (\<lbrakk>m1 = Map.empty; m2 = Map.empty\<rbrakk> \<Longrightarrow> R) \<Longrightarrow> R"
  by (metis map_add_None)

inductive perm :: "'a list \<Rightarrow> 'a list \<Rightarrow> bool" where
perm_base[simp]: "\<lbrakk>bs = []\<rbrakk> \<Longrightarrow> perm [] bs" |
perm_rec:        "\<lbrakk>bs = bs1 @ a # bs2; perm as (bs1 @ bs2)\<rbrakk> \<Longrightarrow> perm (a#as) bs"

inductive_cases perm_Nil[elim!]:  "perm [] bs"
inductive_cases perm_Cons[elim!]: "perm (a#as) bs"

lemma perm_map_distinct:
"\<lbrakk>set xs = set ys; distinct xs; distinct ys\<rbrakk>
   \<Longrightarrow> perm (map f xs) (map f ys)"
proof (induction xs arbitrary: ys)
  case Nil
  then show ?case by simp
next
  case (Cons a xs)
  from `set (a # xs) = set ys` have
    "a \<in> set ys" by auto
  then obtain ys1 ys2 where
    split: "ys = ys1 @ a # ys2" by (meson split_list)
  with `distinct ys` have
    a: "a \<notin> set ys1" "a \<notin> set ys2" by auto

  from `distinct (a#xs)` have
    "distinct xs" by simp
  moreover from `distinct ys` split have
    "distinct (ys1@ys2)" by simp
  moreover from `distinct (a#xs)` `set (a#xs) = set ys` split a have
    "set xs = set (ys1 @ ys2)" by (simp add: insert_eq_iff)
  ultimately have
    hyp: "perm (map f xs) (map f (ys1@ys2))" using Cons.IH by blast
  with split show ?case by (auto intro: perm_rec)
qed

corollary perm_distinct:
"\<lbrakk>set xs = set ys; distinct xs; distinct ys\<rbrakk>
   \<Longrightarrow> perm xs ys"
  using perm_map_distinct[where ?f="id"] by simp

lemma perm_map_remdups:
"\<lbrakk>set xs = set ys\<rbrakk>
   \<Longrightarrow> perm (map f (remdups xs)) (map f (remdups ys))"
  by (auto intro: perm_map_distinct)

context semantics
begin

definition sep :: "'a heap \<Rightarrow> 'a heap \<Rightarrow> 'a heap \<Rightarrow> bool" where
"sep h h1 h2 \<equiv> dom h1 \<inter> dom h2 = {} \<and> h = h1 ++ h2"

lemma sep_com: "sep h h1 h2 = sep h h2 h1"
  unfolding sep_def using map_add_comm by blast
lemma sep_com_intro[intro]: "sep h h1 h2 \<Longrightarrow> sep h h2 h1"
  using sep_com by blast

lemma
[simp]: "sep Map.empty h1 h2 = (Map.empty = h1 \<and> Map.empty = h2)" and
[simp]: "sep h h1 Map.empty = (h = h1)" and
[simp]: "sep h Map.empty h2 = (h = h2)"
  unfolding sep_def by auto

lemma sep_ass[intro]:
"\<lbrakk>sep h h1 h2; sep h1 ha hb\<rbrakk> \<Longrightarrow> sep h ha (hb ++ h2)"
  unfolding sep_def by auto

lemma pred_eq:
  assumes "\<And> s h s' h'. P (s,h) (s',h') = Q (s,h) (s',h')"
  shows   "(P :: 'a pred) = Q"
proof -
  have "\<And> sh sh'. P sh sh' = Q sh sh'" using assms by auto
  then have "\<And> sh. P sh = Q sh" ..
  then show ?thesis ..
qed

(* basic SL connectives *)

fun pure :: "'a pure \<Rightarrow> 'a rel" where
"pure b s s' = (b s \<and> b s')"

fun rel :: "'a rel \<Rightarrow> 'a pred" where
"rel b (s,h) (s',h') = (b s s' \<and> h = Map.empty \<and> h' = Map.empty)"

fun rimp :: "'a rel \<Rightarrow> 'a rel \<Rightarrow> 'a rel" where
"rimp p q s s' = (p s s' \<longrightarrow> q s s')"

fun imp :: "'a pred \<Rightarrow> 'a pred \<Rightarrow> 'a pred" where
"imp p q sh sh' = (p sh sh' \<longrightarrow> q sh sh')"

fun afalse :: "'a pred" where
"afalse st st' = False"

inductive ex :: "var \<Rightarrow> 'a pred \<Rightarrow> 'a pred" where
ex_intro:
"p (s(x:=v),h) (s'(x:=v'),h') \<Longrightarrow> ex x p (s,h) (s',h')"
inductive_cases
ex_elim: "ex b p (s,h) (s',h')"

inductive
sec :: "'level \<Rightarrow> 'a val \<Rightarrow> 'a val \<Rightarrow> 'a rel" where
sec_intro:
  "\<lbrakk>l s = l s'; to_level (l s) \<le> lev \<Longrightarrow> e s = e s'\<rbrakk> \<Longrightarrow> sec lev e l s s'"
inductive_cases
sec_elim: "sec lev e l s s'"

lemma sec_expr_low[simp]:
"sec lev (expr_sem e) (expr_sem Low) s s'
  = (expr_sem e s = expr_sem e s')"
  unfolding sec.simps by auto

fun emp :: "'a pred" where
  "emp (s,h) (s',h') = (h = Map.empty \<and> h' = Map.empty)"

fun const :: "'a heap \<Rightarrow> 'a heap \<Rightarrow> 'a pred" where
  "const H H' (s,h) (s',h') = (h = H \<and> h' = H')"

lemma
true_emp[simp]: "rel (pure (\<lambda> s. True)) = emp" and
const_emp[simp]: "const Map.empty Map.empty = emp"
  by (auto intro!: pred_eq)

inductive
pto :: "'a val \<Rightarrow> 'a val \<Rightarrow> 'a pred" where
pto_intro:
  "pto p v (s,[p s \<mapsto> v s]) (s',[p s' \<mapsto> v s'])"
inductive_cases
pto_elim: "pto p v (s,h) (s',h')"

(* separating conjunction *)

inductive
star :: "'a pred \<Rightarrow> 'a pred \<Rightarrow> 'a pred" where
star_intro:
  "\<lbrakk>sep h h1 h2; sep h' h1' h2'; P (s,h1) (s',h1'); Q (s,h2) (s',h2')\<rbrakk>
    \<Longrightarrow> star P Q (s,h) (s',h')"
inductive_cases
star_elim: "star P Q (s,h) (s',h')" (* don't automatically do this *)

lemma
star_emp_left [simp]: "(star emp P) = P" and
star_emp_right[simp]: "(star P emp) = P"
  by (auto intro!: pred_eq star_intro elim!: star_elim)

lemma
star_rel_left[simp]:  "star (rel t) P (s,h) (s',h') = ((rel t) (s,Map.empty) (s',Map.empty) \<and> P (s,h) (s',h'))" 
  by (auto intro!: star_intro sec_intro elim!: star_elim sec_elim)

lemma star_com_intro: "star P Q (s,h) (s',h') \<Longrightarrow> star Q P (s,h) (s',h')"
proof (erule star_elim)
  fix h1 h2 h1' h2'

  assume "sep h h1 h2" "sep h' h1' h2'"
  then have sep: "sep h h2 h1" "sep h' h2' h1'"
    by auto

  assume "P (s, h1) (s', h1')" "Q (s, h2) (s', h2')"
  with sep show "star Q P (s,h) (s',h')"
    by (auto intro: star_intro)
qed

corollary star_com: "star P Q = star Q P"
  by (auto intro: pred_eq star_com_intro)

lemma star_ass_left: "star (star P Q) R (s,h) (s',h') \<Longrightarrow> star P (star Q R) (s,h) (s',h')"
proof (auto elim!: star_elim)
  fix hp hp' hq hq' hr hr' hpq hpq'

  assume "sep h hpq hr"  "sep h' hpq' hr'"
         "sep hpq hp hq" "sep hpq' hp' hq'"
  then have sep:
         "sep (hq ++ hr) hq hr" "sep (hq' ++ hr') hq' hr'"
         "sep h hp (hq ++ hr)" "sep h' hp' (hq' ++ hr')"
    unfolding sep_def by auto

  assume "P (s, hp) (s', hp')" "Q (s, hq) (s', hq')" "R (s, hr) (s', hr')"
  with sep show "star P (star Q R) (s, h) (s', h')"
    by (auto intro: star_intro)
qed

lemma star_ass_right: "star P (star Q R) (s,h) (s',h') \<Longrightarrow> star (star P Q) R (s,h) (s',h')"
proof (auto elim!: star_elim)
  fix hp hp' hq hq' hr hr' hqr hqr'

  assume "sep h hp hqr"  "sep h' hp' hqr'"
         "sep hqr hq hr" "sep hqr' hq' hr'"
  then have sep:
         "sep (hp ++ hq) hp hq" "sep (hp' ++ hq') hp' hq'"
         "sep h (hp ++ hq) hr" "sep h' (hp' ++ hq') hr'"
    unfolding sep_def by auto

  assume "P (s, hp) (s', hp')" "Q (s, hq) (s', hq')" "R (s, hr) (s', hr')"
  with sep show "star (star P Q) R (s,h) (s',h')"
    by (auto intro: star_intro)
qed

corollary star_ass[simp]: "star (star P Q) R = star P (star Q R)"
  by (auto intro: pred_eq star_ass_left star_ass_right)

lemma star_com12_: "star P (star Q A) = star Q (star P A)" using star_ass[of P Q A] star_com[of P Q] by simp
lemma star_com23:  "star A (star P Q) = star A (star Q P)" using star_com[of P Q] by simp
lemma star_com23_: "star A (star P (star Q B)) = star A (star Q (star P B))" using star_com12_[of P Q B] by simp

lemma star_ac: "(star P (star Q R)) = (star Q (star P R))"
  using star_com star_ass by simp

fun stars :: "('a pred) list \<Rightarrow> 'a pred" where
stars_Nil:  "stars [] = emp" |
stars_Cons: "stars (p#ps) = star p (stars ps)"

lemma stars_app[simp]:
"stars (ps@qs) = star (stars ps) (stars qs)"
  by (induction ps, auto)

lemma stars_reorder_perm:
"perm ps qs
  \<Longrightarrow> stars ps = stars qs"
proof (induction rule: perm.induct)
  case perm_base
  then show ?case by simp
next
  case (perm_rec bs bs1 a bs2 as)
  note stuff = this
  have "star a (star (stars bs1) (stars bs2))
                 = star (star a (stars bs1)) (stars bs2)" by simp
  also have "... = star (star (stars bs1) a) (stars bs2)" by (auto simp: star_com)
  also have "... = star (stars bs1) (star a (stars bs2))" by simp
  finally show ?case using stuff by simp
qed

corollary stars_reorder_distinct:
"\<lbrakk>set ps = set qs; distinct ps; distinct qs\<rbrakk>
  \<Longrightarrow> stars ps = stars qs"
  by (auto intro: stars_reorder_perm perm_distinct)

(* lifting to syntax *)
                                   
primrec rexpr_sem :: "'level \<Rightarrow> rexpr \<Rightarrow> 'a rel" where
"rexpr_sem lev (Pure e) = pure (pure_sem e)" |
"rexpr_sem lev (Sec e l) = sec lev (expr_sem e) (expr_sem l)" |
"rexpr_sem lev (RImp e f) = rimp (rexpr_sem lev e) (rexpr_sem lev f)"

definition
  PTrue :: expr where "PTrue \<equiv> (Eq (App ''low'' []) (App ''low'' []))"

lemma PTrue_sem[simp]:
  "pure_sem PTrue s" by (simp add: PTrue_def)

definition
  PFalse :: expr where "PFalse \<equiv> Not PTrue"

lemma PFalse_sem[simp]:
  "\<not> pure_sem PFalse s" by (simp add: PFalse_def)

definition RNot :: "rexpr \<Rightarrow> rexpr" where
  "RNot e \<equiv> (RImp e (Pure PFalse))"

lemma RNot_sem[simp]:
  "rexpr_sem lev (RNot e) s s' = (\<not> rexpr_sem lev e s s')"
  by (simp add: RNot_def)

definition ROr :: "rexpr \<Rightarrow> rexpr \<Rightarrow> rexpr" where
  "ROr e f \<equiv> (RImp (RNot e) f)"

lemma ROr_sem[simp]:
  "rexpr_sem lev (ROr e f) s s' = (rexpr_sem lev e s s' \<or> rexpr_sem lev f s s')"
  by (auto simp: ROr_def)

definition RAnd :: "rexpr \<Rightarrow> rexpr \<Rightarrow> rexpr" where
  "RAnd e f \<equiv> (RNot (ROr (RNot e) (RNot f)))"

lemma RAnd_sem[simp]:
  "rexpr_sem lev (RAnd e f) s s' = (rexpr_sem lev e s s' \<and> rexpr_sem lev f s s')"
  by (auto simp: RAnd_def)

lemma RImp_sem:
  "rexpr_sem lev (RImp e f) s s' = (rexpr_sem lev e s s' \<longrightarrow> rexpr_sem lev f s s')"
  by simp

primrec assert_sem  :: "'level \<Rightarrow> assert \<Rightarrow> 'a pred" where
"assert_sem lev Emp          = emp" |
"assert_sem lev (Rel p)      = rel (rexpr_sem lev p)" |
"assert_sem lev (Pto p e)    = pto (expr_sem p) (expr_sem e)" |
"assert_sem lev (Star p q)   = star (assert_sem lev p) (assert_sem lev q)" |
"assert_sem lev (Ex x p)     = ex x (assert_sem lev p)" |
"assert_sem lev (AImp p q)   = imp (assert_sem lev p) (assert_sem lev q)" |
"assert_sem lev AFalse       = afalse"


definition
  ANot :: "assert \<Rightarrow> assert" where
  "ANot p \<equiv> (AImp p AFalse)"

lemma ANot_sem[simp]:
  "assert_sem lev (ANot p) st st' = (\<not> (assert_sem lev p st st'))"
  by(simp add: ANot_def)

definition AOr :: "assert \<Rightarrow> assert \<Rightarrow> assert" where
  "AOr p q \<equiv> (AImp (ANot p) q)"

lemma AOr_sem[simp]:
  "assert_sem lev (AOr p q) st st' = ((assert_sem lev p st st') \<or> (assert_sem lev q st st'))"
  by(auto simp add: AOr_def)

definition AAnd :: "assert \<Rightarrow> assert \<Rightarrow> assert" where
  "AAnd e f \<equiv> (ANot (AOr (ANot e) (ANot f)))"

lemma AAnd_sem[simp]:
  "assert_sem lev (AAnd e f) s s' = (assert_sem lev e s s' \<and> assert_sem lev f s s')"
  by (auto simp: AAnd_def)

lemma AImp_sem:
  "assert_sem lev (AImp e f) s s' = (assert_sem lev e s s' \<longrightarrow> assert_sem lev f s s')"
  by simp


(* Let us support forall in the logic *)
definition AForall :: "var \<Rightarrow> assert \<Rightarrow> assert" where
"AForall x p \<equiv> (ANot (Ex x (ANot p)))"

lemma AForall_sem[simp]:
  "assert_sem lev (AForall x p) (s,h) (s',h') = (\<forall>v v'. assert_sem lev p (s(x:=v),h) (s'(x:=v'),h'))"
  by(simp add: AForall_def ex.simps)

lemma Stars_sem[simp]:
"assert_sem lev (Stars ps) = stars (map (assert_sem lev) ps)"
  by (induction ps, auto)

(* heap access *)

lemma
(* the first two loop badly *)
pto_read:  "pto p e (s,h) (s',h') \<Longrightarrow> h (p s) = Some (e s)" and
pto_read': "pto p e (s,h) (s',h') \<Longrightarrow> h' (p s') = Some (e s')" and
pto_frame_read[simp]:  "star (pto p e) r (s,h) (s',h') \<Longrightarrow> h (p s) = Some (e s)" and
pto_frame_read'[simp]: "star (pto p e) r (s,h) (s',h') \<Longrightarrow> h' (p s') = Some (e s')"
  by (auto simp: sep_def map_add_Some_iff elim!: star_elim pto_elim)

lemma pto_write:
"pto p e (s,h) (s',h') \<Longrightarrow> pto p e' (s,h(p s \<mapsto> e' s)) (s',h'(p s' \<mapsto> e' s'))"
  by (auto elim!: pto_elim intro!: pto_intro)

lemma pto_frame_write:
"star (pto p e) r (s,h) (s',h')
  \<Longrightarrow> star (pto p e') r (s,h(p s \<mapsto> e' s)) (s',h'(p s' \<mapsto> e' s'))"
proof -
  assume "star (pto p e) r (s,h) (s',h')"
  then obtain h1 h2 h1' h2' where
    "sep h h1 h2" "sep h' h1' h2'"
    "pto p e (s,h1) (s',h1')"
    "r (s, h2) (s', h2')"
    by (auto elim: star_elim)
  then have
    "sep (h(p s \<mapsto> e' s)) (h1(p s \<mapsto> e' s)) h2"
    "sep (h'(p s' \<mapsto> e' s')) (h1'(p s' \<mapsto> e' s')) h2'"
    "pto p e' (s, h1(p s \<mapsto> e' s)) (s', h1'(p s' \<mapsto> e' s'))"
    "r (s, h2) (s', h2')"
    by (auto simp: sep_def domIff map_add_upd_left intro: pto_intro elim!: pto_elim)
  then show "star (pto p e') r (s,h(p s \<mapsto> e' s)) (s',h'(p s' \<mapsto> e' s'))" ..
qed

(* some more functions *)

definition implies :: "'a pred \<Rightarrow> 'a pred \<Rightarrow> bool" where
"implies P Q \<equiv> \<forall> s h s' h'. P (s,h) (s',h') \<longrightarrow> Q (s,h) (s',h')"

lemma
implies_refl[simp]:   "implies P P" and
implies_trans[intro]: "implies P Q \<Longrightarrow> implies Q R \<Longrightarrow> implies P R"
  unfolding implies_def by blast+

lemma implies_conseq[intro]:
"\<lbrakk>implies P Q; P (s,h) (s',h')\<rbrakk> \<Longrightarrow> Q (s,h) (s',h')"
  unfolding implies_def by blast

lemma implies_frame[intro]: (* frame is typically on the right hand side of the star *)
"implies P Q \<Longrightarrow> implies (star P R) (star Q R)"
  unfolding implies_def by (auto intro: star_intro elim!: star_elim)

lemma
star_mono:       "\<lbrakk>implies P P'; implies Q Q'; star P Q (s,h) (s',h')\<rbrakk> \<Longrightarrow> star P' Q' (s,h) (s',h')" and
star_mono_left:  "\<lbrakk>implies P P'; star P Q (s,h) (s',h')\<rbrakk> \<Longrightarrow> star P' Q (s,h) (s',h')" and
star_mono_right: "\<lbrakk>implies Q Q'; star P Q (s,h) (s',h')\<rbrakk> \<Longrightarrow> star P Q' (s,h) (s',h')"
  by (auto intro: star_intro elim!: star_elim)

(* symmetry of assertions wrt. the two states *)

definition pred_sym :: "'a pred \<Rightarrow> bool" where
"pred_sym P \<equiv> \<forall> s h s' h'. P (s,h) (s',h') = P (s',h') (s,h)"

lemma pred_sym_intro:
"(\<And> s h s' h'. P (s,h) (s',h') \<Longrightarrow> P (s',h') (s,h)) \<Longrightarrow> pred_sym P"
  unfolding pred_sym_def by blast

lemma pred_sym_swap:
"\<lbrakk>pred_sym P; P (s,h) (s',h')\<rbrakk> \<Longrightarrow> P (s',h') (s,h)"
  unfolding pred_sym_def by blast

lemma rexpr_sem_sym[intro]: "rexpr_sem lev a s s' \<Longrightarrow> rexpr_sem lev a s' s"
  by(induction a arbitrary: s s',
     auto elim!: sec.cases intro: sec_intro)

lemma assert_sem_sym[intro]: "pred_sym (assert_sem lev a)"
  by (induction a)
     (auto dest: pred_sym_swap intro!: star_intro ex_intro pred_sym_intro elim!: star_elim ex_elim simp: sec.simps pto.simps)

(* coincidence wrt. a single non-free variable (used in assignments) *)

lemma emp_coincidence[simp]:
"emp (s(x:=v),h) (s'(x:=v'),h') = emp (s,h) (s',h')"
  by auto
lemma pure_coincidence[simp]:
"x \<notin> expr_free b \<Longrightarrow> pure (pure_sem b) (s(x:=v)) (s'(x:=v')) = pure (pure_sem b) s s'"
  by auto
lemma sec_coincidence[simp]:
"x \<notin> expr_free e \<Longrightarrow> x \<notin> expr_free l \<Longrightarrow> sec lev (expr_sem e) (expr_sem l) (s(x:=v)) (s'(x:=v')) = sec lev (expr_sem e) (expr_sem l) s s'"
  by (auto intro: sec_intro elim!: sec_elim)
lemma pto_coincidence[simp]:
"x \<notin> expr_free p \<Longrightarrow> x \<notin> expr_free e \<Longrightarrow> pto (expr_sem p) (expr_sem e) (s(x:=v),h) (s'(x:=v'),h') = pto (expr_sem p) (expr_sem e) (s,h) (s',h')"
  by (auto simp: pto.simps)
lemma star_coincidence:
"\<lbrakk>\<And> h h'. P (s1, h) (s1', h') = P (s2, h) (s2', h');
  \<And> h h'. Q (s1, h) (s1', h') = Q (s2, h) (s2', h')\<rbrakk>
  \<Longrightarrow> star P Q (s1, h) (s1', h') = star P Q (s2, h) (s2', h')"
  by (auto intro: star_intro elim!: star_elim)
lemma star_coincidence_intro:
"\<lbrakk>\<And> h h'. P (s1, h) (s1', h') = P (s2, h) (s2', h');
  \<And> h h'. Q (s1, h) (s1', h') = Q (s2, h) (s2', h');
  star P Q (s1, h) (s1', h')\<rbrakk>
  \<Longrightarrow> star P Q (s2, h) (s2', h')"
  by (auto intro: star_intro elim!: star_elim)
lemma ex_coincidence_same[simp]:
"ex x (assert_sem lev p) (s(x:=v),h) (s'(x:=v'),h') = ex x (assert_sem lev p) (s,h) (s',h')"
  by (auto intro!: ex_intro elim!: ex_elim)

lemma rexpr_sem_coincidence[simp]:
  "x \<notin> rexpr_free p \<Longrightarrow> rexpr_sem lev p (s(x := v)) (s'(x := v')) = rexpr_sem lev p s s'"
  by(induction p, auto)

lemma assert_coincidence[simp]:
"x \<notin> assert_free a \<Longrightarrow> assert_sem lev a (s(x:=v),h) (s'(x:=v'),h') = assert_sem lev a (s,h) (s',h')"
  apply (induction a arbitrary: h h' s s')
  apply (auto intro!: star_coincidence)
  by (elim ex_elim;
      rename_tac va vb, rule_tac v=va and v'=vb in ex_intro, 
      metis fun_upd_twist fun_upd_upd)+

(* coincidence wrt. an entire set of non-free variable (used in framing) *)

lemma pure_coincidence_strong:
"\<lbrakk>coincide s1 s2 (expr_free b); coincide s1' s2' (expr_free b)\<rbrakk>
  \<Longrightarrow> pure (pure_sem b) s2 s2' = pure (pure_sem b) s1 s1'"
  by auto
lemma sec_coincidence_strong:
"\<lbrakk>coincide s1 s2 (expr_free e); coincide s1' s2' (expr_free e);
  coincide s1 s2 (expr_free l); coincide s1' s2' (expr_free l)\<rbrakk>
  \<Longrightarrow> sec lev (expr_sem e) (expr_sem l) s2 s2' = sec lev (expr_sem e) (expr_sem l) s1 s1'"
  by (auto simp: sec.simps)
lemma rexpr_sem_coincidence_strong:
"\<lbrakk>coincide s1 s2 (rexpr_free p); coincide s1' s2' (rexpr_free p)\<rbrakk>
  \<Longrightarrow> rexpr_sem lev p s2 s2' = rexpr_sem lev p s1 s1'"
  by(induction p, auto simp: sec_coincidence_strong)


lemma pto_coincidence_strong:
"\<lbrakk>coincide s1 s2 (expr_free p); coincide s1' s2' (expr_free p);
  coincide s1 s2 (expr_free e); coincide s1' s2' (expr_free e)\<rbrakk>
  \<Longrightarrow> pto (expr_sem p) (expr_sem e) (s2,h) (s2',h') = pto (expr_sem p) (expr_sem e) (s1,h) (s1',h')"
  by (auto simp: pto.simps)

lemma coincide_strengthen[simp]:
  "coincide s1 s2 (P - {x}) \<Longrightarrow> coincide (s1(x := v)) (s2(x := v)) P"
  by (simp add: coincide_reduce)

lemma assert_coincidence_strong:
"\<lbrakk>coincide s1 s2 (assert_free a); coincide s1' s2' (assert_free a)\<rbrakk>
  \<Longrightarrow> assert_sem lev a (s2,h) (s2',h') = assert_sem lev a (s1,h) (s1',h')"
  apply (induction a arbitrary: h h' s1 s2 s1' s2')
        apply (auto intro!: star_coincidence simp: rexpr_sem_coincidence_strong  pure_coincidence_strong pto_coincidence_strong sec_coincidence_strong elim!: ex_elim,
               (metis (full_types) ex_intro coincide_strengthen)+)
  done

lemma coincide_sep_subset_left:
  assumes "sep h h1 h2" and "sep h' h1' h2'"
  assumes "xs \<subseteq> dom h1" "xs \<subseteq> dom h1'"
  shows   "coincide h h' xs = coincide h1 h1' xs"
  using assms unfolding coincide_def sep_def
  by (metis map_add_comm map_add_dom_app_simps(1) set_rev_mp)

lemma coincide_sep_subset_right:
  assumes "sep h h1 h2" and "sep h' h1' h2'"
  assumes "xs \<subseteq> dom h2" "xs \<subseteq> dom h2'"
  shows   "coincide h h' xs = coincide h2 h2' xs"
  using assms unfolding coincide_def sep_def
  by (metis map_le_def map_le_map_add subset_eq)

lemma coincide_sep:
  assumes sep: "sep h h1 h2" "sep h' h1' h2'"
  assumes dom1: "xs \<subseteq> dom h1" and dom2: "ys \<subseteq> dom h2" 
  assumes co1: "coincide h1 h1' xs"
  assumes co2: "coincide h2 h2' ys"
  shows        "coincide h h' (xs \<union> ys)"
proof -
  have "coincide h h' xs"
    unfolding coincide_def proof (clarify)
    fix x
    assume "x \<in> xs"
    with dom1 co1 have "h1 x = h1' x" "x \<in> dom h1" unfolding coincide_def by blast+
    with sep show "h x = h' x"
      by (metis domD map_add_comm map_add_find_right sep_def) 
  qed
  moreover have "coincide h h' ys"
    unfolding coincide_def proof (clarify)
    fix x
    assume "x \<in> ys"
    with dom2 co2 have "h2 x = h2' x" "x \<in> dom h2"
      unfolding coincide_def by blast+
    with sep show "h x = h' x"
      by (metis domD map_add_find_right sep_def) 
  qed
  ultimately show ?thesis by simp
qed


end

end
