# Verdeca

A prototype program verification tool for proving information flow security
and secure declassification of concurrent C code.

Includes case study programs from the paper
"Assume But Verify: Deductive Verification of Leaked Information in
Concurrent Applications".

This implementation was built by extending the SecC verifier.

## Requirements

* Java (tested with Java version 18)

## Supported Platforms

* Linux (tested with Ubuntu 16.04 and 18.04)
* Mac OS (tested with 10.13 and 10.14)

## Installation

### Java

First check that Java is installed. Run `java -version` and look for
a version number that begins with `18`.


## Building Verdeca

Verdeca should build by simply typing `make` in the top-level directory.

```
make
```

This should produce a shell script `Verdeca.sh` for running Verdeca.

## Running Verdeca

Simply run `Verdeca.sh`, supplying a list of files to analyse as command line
arguments.

```
./Verdeca.sh file1.c file2.c ...
```


## Case Studies

We provide instruction for how to check the proof (ie. re-run the
verification) of each case study. Additionally we provide instructions for
how to build and run each of the case studies from Section 7. Doing so
will require a C compiler. Tested with clang. 

### Average Example (Figure 1 and Section 2)

```
./Verdeca.sh examples/case-studies/average.c
```

Should produce output showing each function being successfully verified:
```
examples/case-studies/average.c
  avg_sum_thread ...   success ❤ (time 162ms)
  avg_declass_thread ...   success ❤ (time 58ms)
```

The declassification policy is defined by the specification function
`safe_to_declassify` whose precondition encodes the policy condition and
whose postcondition encodes the release formula.

### Location Service (Section 7.1)

```
./Verdeca.sh examples/case-studies/location-server.c
```

Likewise should produce output showing each function being successfully
verified.

This declassification policy is checked in-line in this example.
This is done on lines 432--438. Lines 433--436 assert (prove) that the
policy condition holds. Line 438 then assumes the release formula.

This example can be built by running

```
make location-main
```

and run by executing `./location-main` which produces output like:

```
Secret location: 0.000000°N, 0.000000°W
Public location: 0.000006°N, 0.000036°W, database length: 1, budget: 0.8
Secret location: 0.007071°N, 0.007071°W
Public location: 0.007101°N, 0.007081°W, database length: 2, budget: 0.6
Secret location: 0.014142°N, 0.014142°W
Public location: 0.014133°N, 0.014144°W, database length: 3, budget: 0.4
...
```

### Auction Server (Section 7.2)

```
./Verdeca.sh examples/case-studies/auction.c
```

Likewise should produce output showing each function being successfully
verified.

The policy condition for this example is defined by the
`safe_to_declassify_bid` predicate (lines 149--155).

The declassification is performed on lines 601--604: the policy condition
is checked on line 601; the declassification (assuming the release
formula) is then done on lines 602--603. Line 604 simply unfolds the
`safe_to_declassify_bid` predicate to ease subsequent reasoning in the
verification.

This example can be built by running
```
make auction-main
```

The auction server can then be run by running './auction-main` passing
an auction duration (in seconds) and an optional reserve price.

In a separate terminal,
the script `examples/simulate-auction-clients.sh` can be used to simulate
a collection of concurrent clients connecting to the auction server,
each after a random delay of a few seconds, and
bidding on the auction. It will produce a series of "OK" results as each
client places its bid.

The auction server will print out the auction result after duration
seconds.

### Wordle (Section 7.3)

``
./Verdeca.sh examples/case-studies/wordle.c
```

Likewise should produce output showing each function being successfully
verified.

The condition for the declassification policy is checked on line
687. The condition is defined by the predicate `safe_to_release` defined
on line 631. The declassification is performed by assuming the release formula
on line 688. The release formula is captured by the
definition `release_policy` defined on line 92.

This case study can be built by running

```
make wordle-main
```

and then run by choosing a 5-letter word e.g. HELLO and executing
`./wordle-main HELLO`

In a separate terminal,
the script `examples/wordle-client.sh` implements a simple guessing client,
e.g. `./examples/wordle-client.sh GUESS`. It will print out the response
using numbers to represent the colours as follows:

0 = BLACK / GREY (incorrect guess)
1 = YELLOW (letter appears in the word but not in this position)
2 = GREEN (letter appears here in the word)

The server can be stopped by typing CTRL-C

### Private Learning (Section 7.4)

```
./Verdeca.sh examples/case-studies/private-learning.c
```

Likewise should produce output showing each function being successfully
verified.

The policy condition for the secure declassification is checked on line
1530. The policy condition is captured by the definition
`safe_to_release_final_m_c` (defined on line 185).
The release formula is assumed on line 1537 (after dereferncing two
pointers to access the data from the heap to be declassified).

This case study can be built by running

```
make learning-main
```

and executed by running `./learning-main` which produces output like:

```
/learning-main 
Trained Parameters: Status code = 1 m = 0.307800 c = 0.397130
Trained Parameters: Status code = 1 m = 0.253659 c = 0.419404
Trained Parameters: Status code = 1 m = 0.208065 c = 0.438162
Trained Parameters: Status code = 1 m = 0.169669 c = 0.453958
Trained Parameters: Status code = 1 m = 0.137334 c = 0.467261
Trained Parameters: Status code = 1 m = 0.110104 c = 0.478463
Trained Parameters: Status code = 1 m = 0.087173 c = 0.487897
Trained Parameters: Status code = 1 m = 0.067862 c = 0.495842
Trained Parameters: Status code = 1 m = 0.051599 c = 0.502532
Trained Parameters: Status code = 1 m = 0.037904 c = 0.508167
Trained Parameters: Status code = 1 m = 0.026371 c = 0.512911
...
```

This program should be stopped by CTRL-C.
